
/* S3 multicore simulator

	S2.1 instruction set with interrupt
   first release						17 Apr 2016
   add: wfi, intx, sync, semaphore		20 Apr 2016

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University

*/

#include "s30.h"

//#define currentp	M[1501]

int R[NC][32], Ir[NC], Pc[NC], M[MAXMEM], RetAds[NC], intflag[NC];
int runflag[NC], savePc[NC], lastInt;
int T;					// global clock
int currentcore;

int signx2( int d ){	// sign bit 21 extended
	if( d & 0x0200000 ) return d | 0xFFC00000;
	return d;
}

void interrupt(int k){
	if( intflag[k] ){
		printf("interrupt core %d\n",k);
		lastInt = T;			// reset int interval
		RetAds[k] = Pc[k];		// save pc
		Pc[k] = M[INTVEC];
		if( Pc[k] < 0 || Pc[k] > 10000)
			error(k,"Pc out of range");
	}
}

void trap( int k, int num ){	// special function
	int a, i;

	switch( num ) {
	case 0:  // stop
		printf("stop, execute %d instructions\n",T);
		runflag[k] = 0;
		break;
	case 1: // print integer
		printf("%d",R[k][1]);
		break;
	case 2: // printc
		printf("%c",R[k][1]);
		break;

//	case 4: dumpreg(); break;
	case 5: R[k][27] = k; break;	// current core

	// number 10..18   support MOS
/*
	case 10: R[27] = newp(); break;
	case 11: R[27] = newStack(); break;
	case 12: enqueue(R[1]); break;
	case 13: terminate(); break;
	case 14: R[27] = nextp(); break;
*/
	case 15: intflag[k] = 0; break;
	case 16: intflag[k] = 1; break;
	case 17: printf("<%d> ",R[k][1]); break;
	case 18: printf("[%d] ",R[k][1]); break;
/*
	case 19:
		if(currentp == 2100)
   			printf("<%d> ",R[1]);
		else
			printf("[%d] ",R[1]);
		break;
*/
	case 20: R[k][27] = newsem(); break;
	case 21: wait(R[k][1]); break;
	case 22: signal(R[k][1]); break;
	case 23:
		RetAds[k] = Pc[k];		// wait for int
		runflag[k] = 0;
		break;
	case 24:					// intx core
		a = R[k][1];			// core number
/*
		if( runflag[a] ){   	// core a is active
			interrupt(a);		// execute int a
		}else{
			runflag[a] = 1;		// intx core a
			Pc[a] = RetAds[a];
		}
*/
		if( runflag[a] == 0 ){	// core a is in wfi
			runflag[a] = 1;
			Pc[a] = RetAds[a];
		}
		break;
	case 25:					// sync all cores
		runflag[k] = 0;			// stop this core
		a = 0;
		for(i=0; i<NC; i++)		// check all cores
			if(runflag[i] == 0) a++;
		if(a == NC)				// all cores have stop
			for(i=0; i<NC; i++)
				runflag[i] = 1;
		break;
	}
}

void run(int k){		// execute one instruction
	int ads, d, r1, r2, r3, i;

	savePc[k] = Pc[k];
	Ir[k] = M[Pc[k]];		// instruction fetch
	Pc[k]++;
	ads = IRads(k);			// decode
	d = IRdisp(k);			// signx  17-bit
	r1 = IRr1(k);
	r2 = IRr2(k);
	r3 = IRr3(k);
	R[k][0] = 0;			// always 0
	switch( IRop(k) ) {
	case LDA:  R[k][r1] = M[ads]; break;
	case LDD:  R[k][r1] = M[R[k][r2]+d]; break;
	case STA:  M[ads] = R[k][r1]; break;
	case STD:  M[R[k][r2]+d] = R[k][r1]; break;
	case MVI:  R[k][r1] = signx2(ads); break;
	case JMP:  Pc[k] = ads; break;
	case JAL:
		R[k][r1] = Pc[k];
		Pc[k] = ads;
		break;
	case JT: if( R[k][r1] != 0 ) Pc[k] = ads; break;
	case JF: if( R[k][r1] == 0 ) Pc[k] = ads; break;
	case ADDI: R[k][r1] = R[k][r2] + d; break;
	case SUBI: R[k][r1] = R[k][r2] - d; break;
	case MULI: R[k][r1] = R[k][r2] * d; break;
	case DIVI: R[k][r1] = R[k][r2] / d; break;
	case ANDI: R[k][r1] = R[k][r2] & d; break;
	case ORI:  R[k][r1] = R[k][r2] | d; break;
	case XORI: R[k][r1] = R[k][r2] ^ d; break;
	case EQI:  R[k][r1] = R[k][r2] == d; break;
	case NEI:  R[k][r1] = R[k][r2] != d; break;
	case LTI:  R[k][r1] = R[k][r2] <  d; break;
	case LEI:  R[k][r1] = R[k][r2] <= d; break;
	case GTI:  R[k][r1] = R[k][r2] >  d; break;
	case GEI:  R[k][r1] = R[k][r2] >= d; break;
	case SHLI: R[k][r1] = R[k][r2] << d; break;
	case SHRI: R[k][r1] = R[k][r2] >> d; break;
	case XOP:
		switch( IRxop(k) ){
		case ADD:  R[k][r1] = R[k][r2] + R[k][r3]; break;
		case SUB:  R[k][r1] = R[k][r2] - R[k][r3]; break;
		case MUL:  R[k][r1] = R[k][r2] * R[k][r3]; break;
		case DIV:  R[k][r1] = R[k][r2] / R[k][r3]; break;
		case AND:  R[k][r1] = R[k][r2] & R[k][r3]; break;
		case OR:   R[k][r1] = R[k][r2] | R[k][r3]; break;
		case XOR:  R[k][r1] = R[k][r2] ^ R[k][r3]; break;
		case EQ:   R[k][r1] = R[k][r2] == R[k][r3]; break;
		case NE:   R[k][r1] = R[k][r2] != R[k][r3]; break;
		case LT:   R[k][r1] = R[k][r2] <  R[k][r3]; break;
		case LE:   R[k][r1] = R[k][r2] <= R[k][r3]; break;
		case GT:   R[k][r1] = R[k][r2] >  R[k][r3]; break;
		case GE:   R[k][r1] = R[k][r2] >= R[k][r3]; break;
		case SHL:  R[k][r1] = R[k][r2] << R[k][r3]; break;
		case SHR:  R[k][r1] = R[k][r2] >> R[k][r3]; break;
		case MV:   R[k][r1] = R[k][r2]; break;
		case LDX:  R[k][r1] = M[R[k][r2] + R[k][r3]]; break;
		case STX:  M[R[k][r2] + R[k][r3]] = R[k][r1]; break;
		case RET:  Pc[k] = R[k][r1]; break;
		case TRAP: trap(k,r1); break;
		case PUSH:
			R[k][r1]++;			// increment stack pointer
			M[R[k][r1]] = R[k][r2];
			break;
		case POP:
			R[k][r2] = M[R[k][r1]];
			R[k][r1]--;			// decrement stack pointer
			break;
		case NOT: R[k][r1] = (R[k][r2] == 0) ? ~0 : 0; break;
		case RETI: Pc[k] = RetAds[k];	break;
		case SAVR:				// push r0..15 to stack
			for(i = 0; i < 16; i++){
				R[k][r1]++;		// sp++
				M[R[k][r1]] = R[k][i];
			}
			break;
		case RESR:				// pop stack to r15..r0
			for(i = 15; i >= 0; i--){
				R[k][i] = M[R[k][r1]];
				R[k][r1]--;		// sp--
			}
			break;
		case SAVT:  R[k][r1] = RetAds[k]; break;
		case REST:  RetAds[k] = R[k][r1]; break;
		case INT:  interrupt(k); break;
		default:
			error(k,"undefine xop");
		}
		break;
	default:
		error(k,"undefine op");
	}
}

PRIVATE int isEI(int k){	// check trap 16, ei instruction
	return (Ir[k] & 0xFFC00FFF) == 0xFC000013;
}

void sim(void){
	int i;
	T++;
	for(i = 0; i < NC; i++){
		currentcore = i;
		if(runflag[i]){
			run(i);       // do not interrupt just after ei
			if(i==0 && (T - lastInt > DEL) && !isEI(i))
				interrupt(0);  // hard interrupt on core 0 only
		}
	}
}

PRIVATE void initcpu(void){
	int i;
	T = 0;
	lastInt = 0;
	for(i = 0; i < NC; i++){
		Pc[i] = 0;
		runflag[i] = 1;
		intflag[i] = 0;
	}
}

int main(int argc, char *argv[]){
	if( argc < 2 ) {
		printf("usage : s30 objfile\n");
		exit(0);
	}
	loadprogram(argv[1]);
	initcpu();
	interp();
	return 0;
}

